/* Module: checkHdr.c

Version  Developer        Date     Change
-------  ---------------  -------  -----------------------
2.1      John Good        28Apr05  Allowed realloc of mHeader string.
2.0      John Good        27Mar05  Added HDU support.
1.10     John Good        21Sep04  That last update was a mistake.  mHeader is
				   sometimes needed externally (e.g. mCoverageCheck)
1.9      John Good        21Sep04  Free mHeader space when no longer needed.
1.8      John Good        11Nov03  Added file existence check if fits
				   open fails.
1.7      John Good        25Aug03  Added status file processing
1.6      John Good        24May03  Added getHdr() call to return header
				   "string". 
1.5.1    A. C. Laity      30Jun03  Added code documentation for fitsCheck
				   and strAdd
1.5      John Good        29Apr03  Added check for failure when file is
				   FITS format but fails FITS open
1.4      John Good        15Apr03  Added special check for DSS headers
                                   (partial; just look for PLTRAH)
1.3      John Good        08Apr03  Also remove <CR> from template lines
1.2      John Good        22Mar03  Renamed wcsCheck to checkWCS for
				   consistency
1.1      John Good        19Mar03  Added limit on template file size
				   (wcsinit() was choking).
1.0      John Good        13Mar03  Baseline code

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "montage.h"
#include "fitsio.h"
#include "wcs.h"

#define MAXHDR 80000

int havePLTRAH;

int haveBITPIX;
int haveNAXIS;
int haveNAXIS1;
int haveNAXIS2;
int haveCTYPE1;
int haveCTYPE2;
int haveCRPIX1;
int haveCRPIX2;
int haveCRVAL1;
int haveCRVAL2;
int haveCDELT1;
int haveCDELT2;
int haveCD1_1;
int haveCD1_2;
int haveCD2_1;
int haveCD2_2;

int debug = 0;

static char *mHeader = (char *)NULL;

int fitsCheck(char *keyword, char *value);
int strAdd(char *header, char *card);
int FITSerror(int status);
int checkWCS(struct WorldCoor *wcs, int action);

char *getHdr();


/***********************************************/
/*                                             */
/*  checkHdr                                   */
/*                                             */
/*  Routine for checking header template files */
/*                                             */
/***********************************************/

int checkHdr(char *infile, int hdrflag, int hdu)
{
   int       i, len, ncard, morekeys;

   int       status = 0;

   char     *keyword;
   char     *value;

   char      fitskeyword[80];
   char      fitsvalue[80];
   char      fitscomment[80];
   char     *end;

   char      line  [1024];
   char      pline [1024];
   FILE     *fp;
   fitsfile *infptr;

   struct WorldCoor *wcs;

   static int maxhdr;

   if(!mHeader)
   {
      mHeader = malloc(MAXHDR);
      maxhdr = MAXHDR;
   }

   havePLTRAH = 0;

   haveBITPIX = 0;
   haveNAXIS  = 0;
   haveNAXIS1 = 0;
   haveNAXIS2 = 0;
   haveCTYPE1 = 0;
   haveCTYPE2 = 0;
   haveCRPIX1 = 0;
   haveCRPIX2 = 0;
   haveCDELT1 = 0;
   haveCDELT2 = 0;
   haveCD1_1  = 0;
   haveCD1_2  = 0;
   haveCD2_1  = 0;
   haveCD2_2  = 0;
   haveCRVAL1 = 0;
   haveCRVAL2 = 0;


   /****************************************/
   /* Initialize the WCS transform library */
   /* and find the pixel location of the   */
   /* sky coordinate specified             */
   /****************************************/

   strcpy(mHeader, "");

   if(fits_open_file(&infptr, infile, READONLY, &status))
   {
      if(debug)
      {
	 printf("\nTemplate file\n");
	 fflush(stdout);
      }

      if(!hdrflag)
      {
	 fp = fopen(infile, "r");

	 if(fp == (FILE *)NULL)
	 {
	    fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"File %s not found.\"]\n", infile);
	    fflush(fstatus);
	    exit(1);
	 }

	 fclose(fp);

	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"File (%s) is not a FITS image\"]\n",
		infile);
	 fflush(fstatus);
	 exit(1);
      }

      fp = fopen(infile, "r");

      if(fp == (FILE *)NULL)
      {
         fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"File %s not found.\"]\n", infile);
         fflush(fstatus);
         exit(1);
      }

      while(1)
      {
         if(fgets(line, 1024, fp) == (char *)NULL)
            break;

         if(line[(int)strlen(line)-1] == '\n')
            line[(int)strlen(line)-1]  = '\0';
	 
         if(line[(int)strlen(line)-1] == '\r')
            line[(int)strlen(line)-1]  = '\0';
	 
	 strcpy(pline, line);

	 if((int)strlen(line) > 80)
	 {
	    fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"FITS header lines cannot be greater than 80 characters.\"]\n");
	    fflush(fstatus);
	    exit(1);
	 }

	 if((int)strlen(mHeader) + 160 < MAXHDR)
         {
	    len = (int)strlen(pline);

	    keyword = pline;

	    while(*keyword == ' ' && keyword < pline+len)
	       ++keyword;

	    end = keyword;

	    while(*end != ' ' && *end != '=' && end < pline+len)
	       ++end;

	    value = end;

	    while((*value == '=' || *value == ' ' || *value == '\'')
		  && value < pline+len)
	       ++value;

	    *end = '\0';
	    end = value;

	    if(*end == '\'')
	       ++end;

	    while(*end != ' ' && *end != '\'' && end < pline+len)
	       ++end;

	    *end = '\0';

            fitsCheck(keyword, value);

	    strAdd(mHeader, line);
	 }
	 else
	 {
	    maxhdr += MAXHDR;
	    mHeader = realloc(mHeader, maxhdr);
	 }
      }

      fclose(fp);
   }
   else
   {
      if(debug)
      {
	 printf("\nFITS file\n");
	 fflush(stdout);
      }

      if(hdrflag)
      {
       fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"FITS file (%s) cannot be used as a header template\"]\n",
		infile);
	 fflush(fstatus);
	 exit(1);
      }

      if(hdu > 0)
      {
         if(fits_movabs_hdu(infptr, hdu+1, NULL, &status))
            FITSerror(status);
      }

      if(fits_get_hdrspace (infptr, &ncard, &morekeys, &status))
	 FITSerror(status);
      
      if(ncard > 1000)
	 mHeader = realloc(mHeader, ncard * 80 + 1024);

      if(debug)
      {
	 printf("ncard = %d\n", ncard);
	 fflush(stdout);
      }

      for (i=1; i<=ncard; i++)
      {
	 if(fits_read_keyn (infptr, i, fitskeyword, fitsvalue, fitscomment, &status))
	    FITSerror(status);

         fitsCheck(fitskeyword, fitsvalue);

	 sprintf(line, "%-8s= %20s", fitskeyword, fitsvalue);

	 if(strncmp(line, "COMMENT", 7) != 0)
	    strAdd(mHeader, line);
      }

      strAdd(mHeader, "END");

      if(fits_close_file(infptr, &status))
         FITSerror(status);
   }



   /********************************************************/
   /*                                                      */
   /* Check to see if we have the minimum FITS header info */
   /*                                                      */
   /********************************************************/

   if(!haveBITPIX)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No BITPIX keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveNAXIS)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No NAXIS keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveNAXIS1)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No NAXIS1 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveNAXIS2)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No NAXIS2 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(havePLTRAH)
   {
      /* If we have this parameter, we'll assume this is a DSS header  */
      /* the WCS checking routine should be able to verify if it isn't */

      free(mHeader);

      maxhdr = 0;
      
      mHeader = (char *)NULL;

      return(0);
   }

   if(!haveCTYPE1)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CTYPE1 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveCTYPE2)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CTYPE2 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveCRPIX1)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CRPIX1 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveCRPIX2)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CRPIX2 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveCRVAL1)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CRVAL1 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveCRVAL2)
   {
      fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CRVAL2 keyword in FITS header\"]\n");
      fflush(fstatus);
      exit(1);
   }

   if(!haveCD1_1 
   || !haveCD1_2 
   || !haveCD2_1 
   || !haveCD2_2)
   {
      if(!haveCDELT1)
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CDELT1 keyword (or incomplete CD matrix) in FITS header\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
      else if(!haveCDELT2)
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"No CDELT2 keyword (or incomplete CD matrix) in FITS header\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }


   /****************************************/
   /* Initialize the WCS transform library */
   /* and find the pixel location of the   */
   /* sky coordinate specified             */
   /****************************************/

   /*
   if(debug)
   {
      printf("header = \n%s\n", mHeader);
      fflush(stdout);
   }
   */

   wcs = wcsinit(mHeader);

   checkWCS(wcs, 0);

   return(0);
}



char *getHdr()
{
   return mHeader;
}



/* fitsCheck checks the value of a given keyword to make */
/* sure it is a valid entry in the FITS header.          */

int fitsCheck(char *keyword, char *value)
{
   char  *end;
   int    ival;
   double dval;

   if(debug)
   {
      printf("fitsCheck() [%s] = [%s]\n", keyword, value);
      fflush(stdout);
   }

   if(strcmp(keyword, "BITPIX") == 0)
   {
      haveBITPIX = 1;

      ival = strtol(value, &end, 0);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"BITPIX keyword in FITS header not an integer\"]\n");
	 fflush(fstatus);
	 exit(1);
      }

      if(ival != 8
      && ival != 16
      && ival != 32
      && ival != 64
      && ival != -32
      && ival != -64)
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"Invalid BITPIX in FITS header (must be 8,16,32,64,-32 or -64)\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "NAXIS") == 0)
   {
      haveNAXIS = 1;

      ival = strtol(value, &end, 0);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"NAXIS keyword in FITS header not an integer\"]\n");
	 fflush(fstatus);
	 exit(1);
      }

      if(ival < 2)
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"NAXIS keyword in FITS header must be >= 2\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "NAXIS1") == 0)
   {
      haveNAXIS1 = 1;

      ival = strtol(value, &end, 0);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"NAXIS1 keyword in FITS header not an integer\"]\n");
	 fflush(fstatus);
	 exit(1);
      }

      if(ival < 0)
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"NAXIS1 keyword in FITS header must be > 0\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "NAXIS2") == 0)
   {
      haveNAXIS2 = 1;

      ival = strtol(value, &end, 0);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"NAXIS2 keyword in FITS header not an integer\"]\n");
	 fflush(fstatus);
	 exit(1);
      }

      if(ival < 0)
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"NAXIS2 keyword in FITS header must be > 0\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "PLTRAH") == 0)
   {
      havePLTRAH = 1;
   }

   else if(strcmp(keyword, "CTYPE1") == 0)
   {
      haveCTYPE1 = 1;
   }

   else if(strcmp(keyword, "CTYPE2") == 0)
   {
      haveCTYPE2 = 1;
   }

   else if(strcmp(keyword, "CRPIX1") == 0)
   {
      haveCRPIX1 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CRPIX1 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CRPIX2") == 0)
   {
      haveCRPIX2 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CRPIX2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CRVAL1") == 0)
   {
      haveCRVAL1 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CRVAL1 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CRVAL2") == 0)
   {
      haveCRVAL2 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CRVAL2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CDELT1") == 0)
   {
      haveCDELT1 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CDELT1 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CDELT2") == 0)
   {
      haveCDELT2 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CDELT2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CROTA2") == 0)
   {
      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CROTA2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CD1_1") == 0)
   {
      haveCD1_1 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CD1_1 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CD1_2") == 0)
   {
      haveCD1_2 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CD1_2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CD2_1") == 0)
   {
      haveCD2_1 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CD1_2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }

   else if(strcmp(keyword, "CD2_2") == 0)
   {
      haveCD2_2 = 1;

      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"CD2_2 keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }
   

   /*
   else if(strcmp(keyword, "EQUINOX") == 0)
   {
      dval = strtod(value, &end);

      if(end < value + (int)strlen(value))
      {
	 fprintf(fstatus, "[struct stat=\"ERROR\", msg=\"EQUINOX keyword in FITS header not a real number\"]\n");
	 fflush(fstatus);
	 exit(1);
      }
   }
   */

   return 0;
}


/* Adds the string "card" to a header line, and */
/* pads the header out to 80 characters.        */

int strAdd(char *header, char *card)
{
   int i;

   int hlen = (int)strlen(header);
   int clen = (int)strlen(card);

   for(i=0; i<clen; ++i)
      header[hlen+i] = card[i];

   if(clen < 80)
      for(i=clen; i<80; ++i)
         header[hlen+i] = ' ';

   header[hlen+80] = '\0';

   return((int)strlen(header));
}



/***********************************/
/*                                 */
/*  Print out FITS library errors  */
/*                                 */
/***********************************/

int FITSerror(int status)
{
   char status_str[FLEN_STATUS];

   fits_get_errstatus(status, status_str);

   fprintf(fstatus, "[struct stat=\"ERROR\", status=%d, msg=\"%s\"]\n", status, status_str);

   exit(1);
}
