/* Module: mMakeImg.c

Version  Developer        Date     Change
-------  ---------------  -------  -----------------------
1.2      John Good        13Oct06  Add 'region' and 'replace' modes
1.1      John Good        13Oct04  Changed format for printing time
1.0      John Good        11Sep03  Baseline code

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "fitsio.h"
#include "coord.h"
#include "wcs.h"
#include "mtbl.h"

#define MAXSTR  256
#define MAXFILE 256


void   printFitsError(int);
void   printError    (char *);
int    readTemplate  (char *filename);
int    parseLine     (char *line);
double ltqnorm       (double);

char  *tval(int);

int  debug;

struct
{
   fitsfile         *fptr;
   long              naxes[2];
   struct WorldCoor *wcs;
   int               sys;
   double            epoch;
}
   output;

static time_t currtime, start;


/******************************************************************/
/*                                                                */
/*  mMakeImg -- A point source image generation program           */
/*                                                                */
/*  A general output FITS image is defined and its pixels are     */
/*  then populated from a table of point sources.  The source     */
/*  fluxes from the table are distributed based on a              */
/*  source-specific point-spread function.                        */
/*                                                                */
/******************************************************************/

int main(int argc, char **argv)
{
   int       i, j, l, m, count, ncol;
   int       dl, dm;
   int       loncol, latcol, fluxcol;
   int       haveTemplate, haveOut, region, replace;
   long      fpixel[4], nelements;
   double    oxpix, oypix;
   int       index, ifile, offscl;
   int       tblSys;
   double    dx, dy, dist2, width2, weight;
   double    noise_level, randnum;
   double    x, y, background;

   double    ilon, ilat;
   double    olon, olat;

   double    pixel_value;
   double    bg1, bg2, bg3, bg4;
   double    noise;

   double  **data;

   int       status = 0;

   char      template_file[MAXSTR];
   char      output_file  [MAXSTR];
   char      table_file   [MAXSTR][MAXFILE];
   char      colname      [MAXSTR][MAXFILE];
   double    width        [MAXFILE];
   double    tblEpoch     [MAXFILE];
   double    refmag       [MAXFILE];

   int       nfile = 0;

   int       bitpix = DOUBLE_IMG; 
   long      naxis  = 2;  


   time(&start);


   /***************************************/
   /* Process the command-line parameters */
   /***************************************/

   if(argc < 3)
   {
      printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
      exit(1);
   }

   debug = 0;

   haveTemplate = 0;
   haveOut      = 0;

   bg1     = 0.;
   bg2     = 0.;
   bg3     = 0.;
   bg4     = 0.;

   noise   = 0.;

   index   = 1;
   region  = 0;
   replace = 0;

   while(1)
   {
      if(index >= argc)
	 break;

      if(argv[index][0] == '-')
      {
	 if(strlen(argv[index]) < 2)
	 {
	    printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
	    exit(1);
	 }

	 switch(argv[index][1])
	 {
	    case 'd':
	       if(argc < index+2)
	       {
		  printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
		  exit(1);
	       }

	       debug = atoi(argv[index+1]);

	       index += 2;

	       break;


	    case 'r':
	       replace = 1;

	       ++index;

	       break;


	    case 'f':
	       region = 1;

	       ++index;

	       break;


	    case 'n':
	       if(argc < index+2)
	       {
		  printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
		  exit(1);
	       }

	       noise = atof(argv[index+1]);

	       index += 2;

	       break;


	    case 'b':
	       if(argc < index+5)
	       {
		  printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
		  exit(1);
	       }

	       bg1 = atof(argv[index+1]);
	       bg2 = atof(argv[index+2]);
	       bg3 = atof(argv[index+3]);
	       bg4 = atof(argv[index+4]);

	       index += 5;

	       break;


	    case 't':
	       if(argc < index+6)
	       {
		  printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
		  exit(1);
	       }

	       strcpy(table_file[nfile], argv[index+1]);
	       strcpy(colname   [nfile], argv[index+2]);

	       width   [nfile] = atof(argv[index+3]);

               if(region)
	          width[nfile] = width[nfile]/3.;
          
	       tblEpoch[nfile] = atof(argv[index+4]);
	       refmag  [nfile] = atof(argv[index+5]);

	       index += 6;

	       ++nfile;

	       break;
	       
	    default:
	       break;
	 }
      }

      else if(!haveTemplate)
      {
	 strcpy(template_file, argv[index]);
	 ++index;
	 haveTemplate = 1;
      }

      else if(!haveOut)
      {
	 strcpy(output_file, argv[index]);
	 ++index;
	 haveOut = 1;
      }

      else
      {
	 printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
	 exit(1);
      }
   }

   if(debug >= 1)
   {
      time(&currtime);
      printf("Done gathering arguments (%.0f seconds)\n", 
	 (double)(currtime - start));
      fflush(stdout);

      printf("debug         = %d\n",  debug);
      printf("noise         = %-g\n", noise);
      printf("bg1           = %-g\n", bg1);
      printf("bg2           = %-g\n", bg2);
      printf("bg3           = %-g\n", bg3);
      printf("bg4           = %-g\n", bg4);
      printf("template_file = %s\n",  template_file);
      printf("output_file   = %s\n",  output_file);
      fflush(stdout);

      for(ifile=0; ifile<nfile; ++ifile)
      {
	 printf("table_file[%d] = %s\n",   ifile, table_file[ifile]);
	 printf("colname   [%d] = %s\n",   ifile, colname   [ifile]);
	 printf("width     [%d] = %-g\n",  ifile, width     [ifile]);
	 printf("tblEpoch  [%d] = %-g\n",  ifile, tblEpoch  [ifile]);
	 printf("refmag    [%d] = %-g\n",  ifile, refmag    [ifile]);
	 fflush(stdout);
      }
   }

   if(!haveOut)
   {
      printf("[struct stat=\"ERROR\", msg=\"Usage: %s [-d level] [-n noise_level] [-b bg1 bg2 bg3 bg4] [-t tbl col width epoch mag] template.hdr out.fits (-t args can be repeated)\"]\n", argv[0]);
      exit(1);
   }

   /*************************************************/ 
   /* Process the output header template to get the */ 
   /* image size, coordinate system and projection  */ 
   /*************************************************/ 

   readTemplate(template_file);

   if(debug >= 1)
   {
      printf("output.naxes[0] =  %ld\n", output.naxes[0]);
      printf("output.naxes[1] =  %ld\n", output.naxes[1]);
      printf("output.sys      =  %d\n",  output.sys);
      printf("output.epoch    =  %-g\n", output.epoch);
      printf("output proj     =  %s\n",  output.wcs->ptype);
      printf("output crval[0] =  %-g\n", output.wcs->crval[0]);
      printf("output crval[1] =  %-g\n", output.wcs->crval[1]);
      printf("output crpix[0] =  %-g\n", output.wcs->crpix[0]);
      printf("output crpix[1] =  %-g\n", output.wcs->crpix[1]);
      printf("output cdelt[0] =  %-g\n", output.wcs->cdelt[0]);
      printf("output cdelt[1] =  %-g\n", output.wcs->cdelt[1]);

      fflush(stdout);
   }


   /***********************************************/ 
   /* Allocate memory for the output image pixels */ 
   /***********************************************/ 

   data = (double **)malloc(output.naxes[1] * sizeof(double *));

   data[0] = (double *)malloc(output.naxes[0] * output.naxes[1] * sizeof(double));

   if(debug >= 1)
   {
      printf("%ld bytes allocated for image pixels\n", 
	 output.naxes[0] * output.naxes[1] * sizeof(double));
      fflush(stdout);
   }


   /**********************************************************/
   /* Initialize pointers to the start of each row of pixels */
   /**********************************************************/

   for(i=1; i<output.naxes[1]; i++)
      data[i] = data[i-1] + output.naxes[0];

   if(debug >= 1)
   {
      printf("pixel line pointers populated\n"); 
      fflush(stdout);
   }


   /**************************************************/
   /* Initialize the data with background plus noise */
   /**************************************************/

   for (j=0; j<output.naxes[1]; ++j)
   {
      for (i=0; i<output.naxes[0]; ++i)
      {
	 x = i / (output.naxes[0] - 1.);
	 y = j / (output.naxes[1] - 1.);

	 background = bg1 * (x * y)
	            + bg2 * y * (1 - x)
	            + bg3 * x * (1 - y)
	            + bg4 * (1 - x) * (1 - y);

	 if(noise > 0.)
	 {
	    randnum = ((double)rand())/RAND_MAX; 

	    noise_level = ltqnorm(randnum) * noise;
	 }
	 else
	    noise_level = 0.;

	 data[j][i] = background + noise_level;

      }
   }


   /************************/
   /* Create the FITS file */
   /************************/

   remove(output_file);               

   if(fits_create_file(&output.fptr, output_file, &status)) 
      printFitsError(status);           


   /*********************************************************/
   /* Create the FITS image.  All the required keywords are */
   /* handled automatically.                                */
   /*********************************************************/

   if (fits_create_img(output.fptr, bitpix, naxis, output.naxes, &status))
      printFitsError(status);          

   if(debug >= 1)
   {
      printf("FITS image created (not yet populated)\n"); 
      fflush(stdout);
   }


   /*****************************/
   /* Loop over the table files */
   /*****************************/

   count = 0;

   for(ifile=0; ifile<nfile; ++ifile)
   {
      /************************************************/ 
      /* Open the table file and find the data column */ 
      /************************************************/ 

      width2 = width[ifile] * width[ifile] / 4.;

      ncol = topen(table_file[ifile]);

      if(ncol <= 0)
      {
	 printf("[struct stat=\"ERROR\", msg=\"Can't open table table %s\"]\n",
	    table_file[ifile]);
	 exit(1);
      }

      tblSys = EQUJ;

      loncol  = tcol("ra");
      latcol  = tcol("dec");

      if(loncol <  0 || latcol <  0)
      {
	 tblSys = GAL;

	 loncol  = tcol("glon");
	 latcol  = tcol("glat");

	 if(loncol <  0 || latcol <  0)
	 {
	    tblSys = ECLJ;

	    loncol  = tcol("elon");
	    latcol  = tcol("elat");

	    if(loncol <  0 || latcol <  0)
	    {
	       printf("[struct stat=\"ERROR\", msg=\"Can't find lon, lat columns\"]\n");

	       exit(1);
	    }
	 }
      }

      fluxcol = tcol(colname[ifile]);

      if(fluxcol <  0)
      {
	 printf("[struct stat=\"ERROR\", msg=\"Can't find flux/magnitude column '%s'\"]\n", colname[ifile]);
	 exit(1);
      }

      /*******************/
      /* For each source */
      /*******************/

      while(tread() >= 0)
      {
	 ++count;

	 ilon = atof(tval(loncol));
	 ilat = atof(tval(latcol));

	 convertCoordinates(    tblSys, tblEpoch[ifile],  ilon,  ilat,
			    output.sys,    output.epoch, &olon, &olat, 0.);

	 pixel_value = atof(tval(fluxcol));

	 if(refmag[ifile] > 0.)
	    pixel_value = pow(10., 0.4 * (refmag[ifile] - pixel_value));

	 wcs2pix(output.wcs, olon, olat, &oxpix, &oypix, &offscl);

	 if(debug >= 3)
	 {
	    printf(" value = %11.3e at coord = (%12.8f,%12.8f) -> (%12.8f,%12.8f)",
	       pixel_value, ilon, ilat, olon, olat);

	    if(offscl)
	    {
	       printf(" -> opix = (%7.1f,%7.1f) OFF SCALE\n",
		  oxpix, oypix);
	       fflush(stdout);
	    }
	    else
	    {
	       printf(" -> opix = (%7.1f,%7.1f)\n",
		  oxpix, oypix);
	       fflush(stdout);
	    }
	 }

	 if(!offscl)
	 {
	    l = (int)(oxpix + 0.5) - 1;
	    m = (int)(oypix + 0.5) - 1;

	    if(l < 0 || m < 0 || l >= output.naxes[0] || m >= output.naxes[1])
	    {
	       printf("[struct stat=\"INFO\", msg=\"Bad Values: l=%d, m=%d\"]\n", l, m);
	       fflush(stdout);
	    }

	    else
	    {
	       for(dl=-3*width[ifile]; dl<=3*width[ifile]; ++dl)
	       {
		  if(l+dl < 0 || l+dl>= output.naxes[0])
		     continue;

		  dx = oxpix - (l+dl);

		  for(dm=-3*width[ifile]; dm<=3*width[ifile]; ++dm)
		  {
		     if(m+dm < 0 || m+dm>= output.naxes[1])
			continue;

		     dy = oypix - (m+dm);

		     dist2 = dx*dx + dy*dy;

                     if(region)
                        weight = 1;
                     else
		        weight = exp(-dist2/width2);

                     
		     if(replace)
                        data[m+dm][l+dl] = weight * pixel_value;
                     else
		        data[m+dm][l+dl] += weight * pixel_value;
		  }
	       }
	    }
	 }
      }

      tclose();
   }

   if(debug >= 1)
   {
      time(&currtime);
      printf("Done processing tables (%.0f seconds)\n", 
	 (double)(currtime - start));
      fflush(stdout);
   }


   /************************/
   /* Write the image data */
   /************************/

   fpixel[0] = 1;
   fpixel[1] = 1;
   fpixel[2] = 1;
   fpixel[3] = 1;
   nelements = output.naxes[0];

   for(j=0; j<output.naxes[1]; ++j)
   {
      if (fits_write_pix(output.fptr, TDOUBLE, fpixel, nelements,
                         (void *)(data[j]), &status))
         printFitsError(status);

      ++fpixel[1];
   }

   if(debug >= 1)
   {
      printf("Data written to FITS data image\n");
      fflush(stdout);
   }


   /*************************************/
   /* Add keywords from a template file */
   /*************************************/

   if(fits_write_key_template(output.fptr, template_file, &status))
      printFitsError(status);           

   if(debug >= 1)
   {
      printf("Template keywords written to FITS image\n"); 
      fflush(stdout);
   }


   /***********************/
   /* Close the FITS file */
   /***********************/

   if(fits_close_file(output.fptr, &status))
      printFitsError(status);           

   if(debug >= 1)
   {
      printf("FITS image finalized\n"); 
      fflush(stdout);
   }

   printf("[struct stat=\"OK\"]\n");
   exit(0);
}



/**************************************************/
/*                                                */
/*  Read the output header template file.         */
/*  Specifically extract the image size info.     */
/*  Also, create a single-string version of the   */
/*  header data and use it to initialize the      */
/*  output WCS transform.                         */
/*                                                */
/**************************************************/

int readTemplate(char *filename)
{
   int       i;

   FILE     *fp;

   char      line[MAXSTR];

   char     *header[2];

   int       sys;
   double    epoch;

   header[0] = malloc(32768);
   header[1] = (char *)NULL;


   /********************************************************/
   /* Open the template file, read and parse all the lines */
   /********************************************************/

   fp = fopen(filename, "r");

   if(fp == (FILE *)NULL)
   {
      printf("[struct stat=\"ERROR\", msg=\"Template file [%s] not found.\"]\n",
         filename);
      exit(1);
   }

   while(1)
   {
      if(fgets(line, MAXSTR, fp) == (char *)NULL)
	 break;

      if(line[strlen(line)-1] == '\n')
         line[strlen(line)-1]  = '\0';

      if(debug >= 2)
      {
	 printf("Template line: [%s]\n", line);
	 fflush(stdout);
      }

      for(i=strlen(line); i<80; ++i)
	 line[i] = ' ';
      
      line[80] = '\0';

      strcat(header[0], line);

      parseLine(line);
   }

   if(debug >= 2)
   {
      printf("\nheader ----------------------------------------\n");
      printf("%s\n", header[0]);
      printf("-----------------------------------------------\n\n");
   }


   /****************************************/
   /* Initialize the WCS transform library */
   /****************************************/

   output.wcs = wcsinit(header[0]);

   if(output.wcs == (struct WorldCoor *)NULL)
   {
      printf("[struct stat=\"ERROR\", msg=\"Output wcsinit() failed. Exiting.\"]\n");
      exit(1);
   }


   /*************************************/
   /*  Set up the coordinate transform  */
   /*************************************/

   if(output.wcs->syswcs == WCS_J2000)
   {
      sys   = EQUJ;
      epoch = 2000.;

      if(output.wcs->equinox == 1950.)
	 epoch = 1950;
   }
   else if(output.wcs->syswcs == WCS_B1950)
   {
      sys   = EQUB;
      epoch = 1950.;

      if(output.wcs->equinox == 2000.)
	 epoch = 2000;
   }
   else if(output.wcs->syswcs == WCS_GALACTIC)
   {
      sys   = GAL;
      epoch = 2000.;
   }
   else if(output.wcs->syswcs == WCS_ECLIPTIC)
   {
      sys   = ECLJ;
      epoch = 2000.;

      if(output.wcs->equinox == 1950.)
      {
	 sys   = ECLB;
	 epoch = 1950.;
      }
   }
   else       
   {
      sys   = EQUJ;
      epoch = 2000.;
   }

   output.sys   = sys;
   output.epoch = epoch;

   free(header[0]);

   return 0;
}



/**************************************************/
/*                                                */
/*  Parse header lines from the template,         */
/*  looking for NAXIS1 and NAXIS2                 */
/*                                                */
/**************************************************/

int parseLine(char *line)
{
   char *keyword;
   char *value;
   char *end;

   int   len;

   len = strlen(line);

   keyword = line;

   while(*keyword == ' ' && keyword < line+len)
      ++keyword;
   
   end = keyword;

   while(*end != ' ' && *end != '=' && end < line+len)
      ++end;

   value = end;

   while((*value == '=' || *value == ' ' || *value == '\'')
         && value < line+len)
      ++value;
   
   *end = '\0';
   end = value;

   if(*end == '\'')
      ++end;

   while(*end != ' ' && *end != '\'' && end < line+len)
      ++end;
   
   *end = '\0';

   if(debug >= 2)
   {
      printf("keyword [%s] = value [%s]\n", keyword, value);
      fflush(stdout);
   }

   if(strcmp(keyword, "NAXIS1") == 0)
      output.naxes[0] = atoi(value);

   if(strcmp(keyword, "NAXIS2") == 0)
      output.naxes[1] = atoi(value);

   return 0;
}


/***********************************/
/*                                 */
/*  Print out FITS library errors  */
/*                                 */
/***********************************/

void printFitsError(int status)
{
   char status_str[FLEN_STATUS];

   fits_get_errstatus(status, status_str);

   printf("[struct stat=\"ERROR\", flag=%d, msg=\"%s\"]\n", status, status_str);

   exit(1);
}
