#!/home/jcg/perl/bin/perl -w
#
# Module: mDAG
#
# Creates an XML-based Directed Acyclic Graph (DAG) that specifies (1) "jobs"
# or Montage modules to be run to produce a mosaic, (2) "files" to be 
# consumed or generated during processing (input, intermediate, and output 
# files), and "dependencies" between jobs and files that imply an order
# in which jobs should be run.  In this "abstract graph", logical names are 
# assigned to jobs and files.  The abstract graph is intended to be an input 
# to a job scheduler, such as Pegasus in the Chimera Virtual Data System.
#
# Inputs:
# 1) Directory name containing the input data files to be mosaicked.
# 2) Montage template header file specifying the mosaic parameters.
# 3) Directory name for output and temporary files produced by this script.
#
# Outputs:
# 1) XML-based DAG
# 2) Data files needed for processing the workflow specified in the DAG:
#    a) Projected image table
#    b) Background-corrected image table
#    c) Table listing the individual difference image fit plane parameter
#       files that will be generated by each run of mFitplane during mosaic
#       processing.
# 
# Version  Developer        Date     Change
# -------  ---------------  -------  -----------------------
# 4.4      Joe Jacob        17Mar05  Allowed more surveys/bands
# 4.3      Joe Jacob        16Mar05  Added survey, width, height arguments
# 4.2      Joe Jacob        08Feb05  Fixed file dependencies in mImgtbl job 
#                                    for the shrunken files
# 4.1      Joe Jacob        09Dec04  Changed tile overlap from 0 to 100
# 4.0      Joe Jacob        30Nov04  Added graphical presentation of output
# 3.3      Joe Jacob        29Nov04  Suffix an optional argument again
# 3.2      Joe Jacob        24Nov04  Added metadata file generation
# 3.1      Joe Jacob        24Nov04  Email argument added for notification
# 3.0      Joe Jacob        24Nov04  Suffix now a required argument
# 2.7      Joe Jacob        03Nov04  Reduced mAdd tile overlap to 0
# 2.6      Joe Jacob        03Nov04  Added -e option to mAdd jobs
# 2.5      Joe Jacob        03Nov04  Changed output mosaic filename
# 2.4      Joe Jacob        03Nov04  Added band selector argument
# 2.3      Joe Jacob        13Oct04  Added support for tiled mImgtbls and mAdds
# 2.2      Joe Jacob        07Oct04  Added mImgtbl job before mAdd
# 2.1      Joe Jacob        07Oct04  Merged mDiff and mFitplane into mDiffFit
# 2.0      Joe Jacob        07Oct04  Implemented DAX schema version 1.8
# 1.3      Joe Jacob        11Dec03  No longer remove image table file
# 1.2      Joe Jacob        10Nov03  Fixed diff area file filenames
# 1.1      Joe Jacob        20Oct03  Added status prints with time stamps
# 1.0      Joe Jacob        09Oct03  Baseline code
# -------  ---------------  -------  -----------------------
#
use File::Basename;
#
# Check syntax.
#
if ((@ARGV != 7) && (@ARGV != 8)) {
    ShowUsage();
    exit 1;
}
#
# Parse command line.
#
$survey = shift;
$MosaicCenter = shift;
$MosaicWidth = shift;
$MosaicHeight = shift;
$band = shift;
$workdir = shift;
$email = shift;
$email =~ /\S+@\S+/ or die "$email not a valid email address: $!";
if (@ARGV) {
    $suffix = shift;
}
else {
    $suffix = "";
}
PrintLog ("Started DAG processing for survey \"$survey\", location \"$MosaicCenter\", width \"$MosaicWidth\", height \"$MosaicHeight\", and band \"$band\"");
#
# Setup. You must specify the directory containing the Montage binaries by 
# setting the MONTAGE environment variable.
#
GetMontageBin (\$MontageVersion, \$mArchiveList, \$mMakeHdr, \$mOverlaps, 
	       \$mDAGTbls, \$mTileHdr, \$mSubset);
$dv_version = "1.0";
$template_fname = "template$suffix.hdr";
$OutPrefix = "${survey}_${MosaicCenter}_${MosaicWidth}_${MosaicHeight}_$band";
$OutPrefix =~ s/\s+/_/g;
$out_fname = "$OutPrefix$suffix.fits";
$out_fname =~ s/\s+/_/g;
$shrunken_out_fname = "shrunken_${out_fname}";
$AreaFname{$shrunken_out_fname} = GetAreaFname($shrunken_out_fname);
$shrunken_jpeg_fname = "shrunken_$OutPrefix$suffix.jpg";
$dag_fname = "adag$suffix.xml";
$img_tbl = "images$suffix.tbl";
$pimg_tbl = "pimages$suffix.tbl";
$cimg_tbl = "cimages$suffix.tbl";
$new_cimg_tbl = "new-$cimg_tbl";
$shrunken_tbl = "simages$suffix.tbl";
$new_shrunken_tbl = "new-$shrunken_tbl";
$shrunken_template_fname = "shrunken$suffix.hdr";
$meta_fname = "Meta.inf";
#
# Make work directory if it doesn't already exist.
#
if (! -e $workdir) {
    mkdir("$workdir") || die "Can't mkdir $workdir: $!";
}
$dag_fname_fullpath = "$workdir/$dag_fname";
$template_fname_fullpath = "$workdir/$template_fname";
$img_tbl_fullpath = "$workdir/$img_tbl";
$pimg_tbl_fullpath = "$workdir/$pimg_tbl";
$cimg_tbl_fullpath = "$workdir/$cimg_tbl";
$shrunken_tbl_fullpath = "$workdir/$shrunken_tbl";
$shrunken_template_fname_fullpath = "$workdir/$shrunken_template_fname";
$meta_fname_fullpath = "$workdir/$meta_fname";
#
# Get table of input files needed for this request.  
#
system "$mArchiveList $survey $band \"$MosaicCenter\" $MosaicWidth $MosaicHeight $img_tbl_fullpath";
PrintLog ("Created input image table with mArchiveList");
#
# Build Montage template header file for this set of input files.
#
system "$mMakeHdr $img_tbl_fullpath $template_fname_fullpath";
PrintLog ("Created Montage template header file with mMakeHdr");
#
# Build the table of projected and background corrected images.
#
system "$mDAGTbls $img_tbl_fullpath $template_fname_fullpath $pimg_tbl_fullpath $cimg_tbl_fullpath";
PrintLog ("Created projected and background corrected image tables with mDAGTbls");
#
# Get list of overlapping images
#
$diffs_tbl_fullpath = "$workdir/diffs$suffix.tbl";
system "$mOverlaps $img_tbl_fullpath $diffs_tbl_fullpath";
PrintLog ("Created list of overlapping images with mOverlaps");
#
# Build a list of input files, a hash linking filename to a file number 
# and another pair of hashes linking filename to projected filename and 
# background corrected filename, respectively.
#
ConstructFileInfo ($img_tbl_fullpath, $pimg_tbl_fullpath, $cimg_tbl_fullpath,
		   \@infiles, \%ProjFname, \%BgCorrFname, \%InFname,
		   \%AreaFname, \%FileNum);
#unlink ("$img_tbl_fullpath") or warn "Couldn't remove file $img_tbl_fullpath: $!";
#
# Build a hash-of-hash flagging overlapping filenames.
#
GetOverlaps ($diffs_tbl_fullpath, \%olaps);
unlink ("$diffs_tbl_fullpath") or warn "Couldn't remove file $diffs_tbl_fullpath: $!";
#
# Determine number of output mosaic tiles.
#
$TileSize = 3600; # Nominal tile width and height in pixels (not including additional padding for tile overlap);
$TileOverlapX = $TileOverlapY = 100; # Number of pixels overlap between tiles
GetImageSize ($template_fname_fullpath, \$naxis1, \$naxis2);
$NumTilesX = sprintf ("%.0f", $naxis1 / $TileSize);
if ($NumTilesX < 1) {
    $NumTilesX = 1;
}
$NumTilesY = sprintf ("%.0f", $naxis2 / $TileSize);
if ($NumTilesY < 1) {
    $NumTilesY = 1;
}
$NumTiles = $NumTilesX * $NumTilesY;
#
# Determine shrink factor for presentation image.
#
$MaxShrunkenImgSize = 1024; # Maximum width or height for presentation image
$ShrinkFactorX = sprintf ("%.0f", $naxis1 / $MaxShrunkenImgSize);
if ($ShrinkFactorX == 0) {
    $ShrinkFactorX = 1;
}
if (($naxis1/$ShrinkFactorX) > $MaxShrunkenImgSize) {
    $ShrinkFactorX++;
}
$ShrinkFactorY = sprintf ("%.0f", $naxis2 / $MaxShrunkenImgSize);
if ($ShrinkFactorY == 0) {
    $ShrinkFactorY = 1;
}
if (($naxis2/$ShrinkFactorY) > $MaxShrunkenImgSize) {
    $ShrinkFactorY++;
}
$ShrinkFactor = ($ShrinkFactorX > $ShrinkFactorY) ? $ShrinkFactorX : $ShrinkFactorY;
#
# Create tiled header files, shrunken template header file, subset image 
# tables, output filenames, shrunken output file names, and shrink factors 
# for tiles.
#
if ($NumTiles > 1) {
    PrintLog ("Using $NumTilesX x $NumTilesY tiles for $naxis1 x $naxis2 image.");
    ShrinkTemplate ($template_fname_fullpath, $ShrinkFactor, 
		    $shrunken_template_fname_fullpath);
    for ($i = 0; $i < $NumTilesX; $i++) {
	for ($j = 0; $j < $NumTilesY; $j++) {
	    $TiledTemplate[$i][$j] = "template${suffix}_${i}_${j}.hdr";
	    $tile_template_fname_fullpath = "$workdir/$TiledTemplate[$i][$j]";
	    $TiledCImgTbl[$i][$j] = "cimages${suffix}_${i}_${j}.tbl";
	    $tile_cimg_tbl_fullpath = "$workdir/$TiledCImgTbl[$i][$j]";
	    $NewTiledCImgTbl[$i][$j] = "new-$TiledCImgTbl[$i][$j]";
	    $TiledOut[$i][$j] = "$OutPrefix${suffix}_${i}_${j}.fits";
	    $ShrunkenTiledOut[$i][$j] = "shrunken_$OutPrefix${suffix}_${i}_${j}.fits";
	    $AreaFname{$TiledOut[$i][$j]} = GetAreaFname($TiledOut[$i][$j]);
	    $AreaFname{$ShrunkenTiledOut[$i][$j]} = GetAreaFname($ShrunkenTiledOut[$i][$j]);
	    system "$mTileHdr $template_fname_fullpath $tile_template_fname_fullpath $NumTilesX $NumTilesY $i $j $TileOverlapX $TileOverlapY";
	    system "$mSubset -f $cimg_tbl_fullpath $tile_template_fname_fullpath $tile_cimg_tbl_fullpath";
	    GetImageSize ($tile_template_fname_fullpath, \$naxis1, \$naxis2);
	    @{$TiledFnames[$i][$j]} = GetFilesFromTable ("$workdir/$TiledCImgTbl[$i][$j]");
	}
    }
    unlink ("$cimg_tbl_fullpath") or warn "Couldn't remove file $cimg_tbl_fullpath: $!";
#
#     Create shrunken file image table.
#
    $Col1Header = "file";
    $Col1Width = length $Col1Header;
    $Col1Type = "char";
    $len = length $Col1Type;
    if ($len > $Col1Width) {
	$Col1Width = $len;
    }
    for ($i = 0; $i < $NumTilesX; $i++) {
	for ($j = 0; $j < $NumTilesY; $j++) {
	    $len = length $ShrunkenTiledOut[$i][$j];
	    if ($len > $Col1Width) {
		$Col1Width = $len;
	    }
	}
    }
    open (SHRUNK_TBL, ">$shrunken_tbl_fullpath") or die "Couldn't open $shrunken_tbl_fullpath for writing: $!";
    print SHRUNK_TBL "\\datatype=fitshdr\n";
    print SHRUNK_TBL "|".PadField($Col1Header, $Col1Width, " ")."|\n";
    print SHRUNK_TBL "|".PadField($Col1Type, $Col1Width, " ")."|\n";
    for ($i = 0; $i < $NumTilesX; $i++) {
	for ($j = 0; $j < $NumTilesY; $j++) {
	    print SHRUNK_TBL " ".PadField($ShrunkenTiledOut[$i][$j], $Col1Width, " ")."\n";
	}
    }
    close (SHRUNK_TBL);
    PrintLog ("Created tiled header files and subset image tables with mTileHdr and mSubset");
}
else {
    $TiledTemplate[0][0] = $template_fname;
    $TiledCImgTbl[0][0] = $cimg_tbl;
    $NewTiledCImgTbl[0][0] = $new_cimg_tbl;
    $TiledOut[0][0] = $out_fname;
    $ShrunkenTiledOut[0][0] = "$shrunken_out_fname";
    $AreaFname{$out_fname} = GetAreaFname($out_fname);
    @{$TiledFnames[0][0]} = GetFilesFromTable ($cimg_tbl_fullpath);
}
#
# Create metadata file
#
$ProxyServerKey = "proxyserver";
$ProxyServer = "birdie.isi.edu";
$ProxyUserKey = "username";
$ProxyUser = "vahi";
$ProxyPasswdKey = "password";
$ProxyPasswd = "demo123";
$SurveyKey = "survey";
$ObjectKey = "object";
$WidthKey = "width";
$HeightKey = "height";
$BandKey = "band";
$EmailKey = "email";
$NumTilesXKey = "nx";
$NumTilesYKey = "ny";
$FilebaseKey = "filebase";
open (META, ">$meta_fname_fullpath") or die "Couldn't open $meta_fname_fullpath for writing: $!";
print META "$ProxyServerKey\t$ProxyServer\n";
print META "$ProxyUserKey\t$ProxyUser\n";
print META "$ProxyPasswdKey\t$ProxyPasswd\n";
print META "$SurveyKey\t$survey\n";
print META "$ObjectKey\t$MosaicCenter\n";
print META "$WidthKey\t$MosaicWidth\n";
print META "$HeightKey\t$MosaicHeight\n";
print META "$BandKey\t$band\n";
print META "$EmailKey\t$email\n";
print META "$NumTilesXKey\t$NumTilesX\n";
print META "$NumTilesYKey\t$NumTilesY\n";
print META "$FilebaseKey\t$OutPrefix\n";
close (META);
#
# Print XML header lines.
#
open (DAG, ">$dag_fname_fullpath") or die "Couldn't open $dag_fname_fullpath for writing: $!";
XML_Header();
#
# Print XML "filename" lines for input images.
#
print DAG "<!-- part 1: list of all files used (may be empty) -->\n";
foreach $infile (@infiles) {
    print DAG "  <filename file=\"$infile\" link=\"input\"/>\n";
}
#
# Print XML "filename" lines for projected images.
#
foreach $infile (@infiles) {
    print DAG "  <filename file=\"$ProjFname{$infile}\" link=\"inout\"/>\n";
}
#
# Print XML "filename" lines for difference images.  Also, build hashes linking
# a pair of filenames to the filename of their difference image and mFitplane 
# output, respectively.  In addition, calculate the width of the columns of the
# fit plane file list to be constructed next.
#
$Col1Header = "cntr1";
$Col1Width = length $Col1Header;
$Col2Header = "cntr2";
$Col2Width = length $Col2Header;
$Col3Header = "stat";
$Col3Width = length $Col3Header;
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	$DiffFname{$fname1}{$fname2} = "diff.$FileNum{$fname1}.$FileNum{$fname2}.fits";
	$AreaFname{$DiffFname{$fname1}{$fname2}} = GetAreaFname($DiffFname{$fname1}{$fname2});
	$FitPlaneFname{$fname1}{$fname2} = "fit.$FileNum{$fname1}.$FileNum{$fname2}.txt";
	$len = length $FileNum{$fname1};
	if ($len > $Col1Width) {
	    $Col1Width = $len;
	}
	$len = length $FileNum{$fname2};
	if ($len > $Col2Width) {
	    $Col2Width = $len;
	}
	$len = length $FitPlaneFname{$fname1}{$fname2};
	if ($len > $Col3Width) {
	    $Col3Width = $len;
	}
	print DAG "  <filename file=\"$DiffFname{$fname1}{$fname2}\" link=\"inout\"/>\n";
    }
}
#
# Create a list of fit plane files to be concatenated during DAG processing.
#
$fit_list = "fit_list$suffix.tbl";
$fit_list_fullpath = "$workdir/$fit_list";
open (FIT_LIST, ">$fit_list_fullpath") or die "Couldn't open $fit_list_fullpath for writing: $!";
print FIT_LIST "|".PadField($Col1Header, $Col1Width, " ")."|".PadField($Col2Header, $Col2Width, " ")."|".PadField($Col3Header, $Col3Width, " ")."|\n";
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	print FIT_LIST " ".PadField("$FileNum{$fname1}", $Col1Width, " ")." ".PadField("$FileNum{$fname2}", $Col2Width, " ")." ".PadField("$FitPlaneFname{$fname1}{$fname2}", $Col3Width, " ")."\n";
    }
}
close (FIT_LIST);
#
# Print XML "filename" lines for mFitPlane output for each difference image
# and the concatenated table of fit planes.
#
$fit_planes = "fits$suffix.tbl";
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	print DAG "  <filename file=\"$FitPlaneFname{$fname1}{$fname2}\" link=\"inout\"/>\n";
    }
}
print DAG "  <filename file=\"$fit_planes\" link=\"inout\"/>\n";
#
# Print XML "filename" line for correction planes output from mBgModel.
#
$correction_planes = "corrections.tbl";
print DAG "  <filename file=\"$correction_planes\" link=\"inout\"/>\n";
#
# Print XML "filename" lines for background corrected images.
#
foreach $infile (@infiles) {
    print DAG "  <filename file=\"$BgCorrFname{$infile}\" link=\"inout\"/>\n";
}
#
# Print XML "filename" lines for area files.
#
foreach $infile (@infiles) {
    print DAG "  <filename file=\"$AreaFname{$ProjFname{$infile}}\" link=\"inout\"/>\n";
    print DAG "  <filename file=\"$AreaFname{$BgCorrFname{$infile}}\" link=\"inout\"/>\n";
}
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	print DAG "  <filename file=\"$AreaFname{$DiffFname{$fname1}{$fname2}}\" link=\"output\"/>\n";
    }
}
#
# Print XML "filename" lines for template files, subset image tables, output
# files and shrunken output files.
#
if ($NumTiles > 1) {
    for ($i = 0; $i < $NumTilesX; $i++) {
	for ($j = 0; $j < $NumTilesY; $j++) {
            print DAG "  <filename file=\"$TiledTemplate[$i][$j]\" link=\"input\"/>\n";
            print DAG "  <filename file=\"$TiledCImgTbl[$i][$j]\" link=\"input\"/>\n";
            print DAG "  <filename file=\"$NewTiledCImgTbl[$i][$j]\" link=\"inout\"/>\n";
	    print DAG "  <filename file=\"$TiledOut[$i][$j]\" link=\"inout\"/>\n";
	    print DAG "  <filename file=\"$AreaFname{$TiledOut[$i][$j]}\" link=\"output\"/>\n";
	    print DAG "  <filename file=\"$ShrunkenTiledOut[$i][$j]\" link=\"inout\"/>\n";
	    print DAG "  <filename file=\"$AreaFname{$ShrunkenTiledOut[$i][$j]}\" link=\"output\"/>\n";
	}
    }
    print DAG "  <filename file=\"$shrunken_template_fname\" link=\"input\"/>\n";
    print DAG "  <filename file=\"$shrunken_tbl\" link=\"input\"/>\n";
    print DAG "  <filename file=\"$new_shrunken_tbl\" link=\"inout\"/>\n";
}
#
# Print remaining XML "filename" lines.
#
print DAG "  <filename file=\"$pimg_tbl\" link=\"input\"/>\n";
print DAG "  <filename file=\"$fit_list\" link=\"input\"/>\n";
print DAG "  <filename file=\"$template_fname\" link=\"input\"/>\n";
print DAG "  <filename file=\"$shrunken_out_fname\" link=\"inout\"/>\n";
print DAG "  <filename file=\"$AreaFname{$shrunken_out_fname}\" link=\"output\"/>\n";
print DAG "  <filename file=\"$shrunken_jpeg_fname\" link=\"output\"/>\n";
#
# Print XML "job" lines for mProject jobs.  Also, build a hash 
# linking filename to job ID for the mProject job.
#
print DAG "<!-- part 2: definition of all jobs (at least one) -->\n";
$job_cnt = 1;
$name = "mProject";
$level = ($NumTiles == 1) ? 9 : 11;
$count = 1;
foreach $infile (@infiles) {
    $jobID = JobNum2ID ($job_cnt++);
    $mProjJob{$infile} = $jobID;
    print DAG "  <job id=\"$jobID\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
    print DAG "    <argument>";
    print DAG "<filename file=\"$infile\"/>";
    print DAG "<filename file=\"$ProjFname{$infile}\"/>";
    print DAG "<filename file=\"$template_fname\"/>";
    print DAG "</argument>\n";
    print DAG "    <uses file=\"$infile\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
    print DAG "    <uses file=\"$ProjFname{$infile}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"$AreaFname{$ProjFname{$infile}}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"$template_fname\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
    print DAG "  </job>\n";
    $count++;
}
#
# Print XML "job" lines for mDiffFit jobs.  Also, build a hash 
# linking a pair of filenames to the job ID for the mDiffFit job.
#
$name = "mDiffFit";
$level--;
$count = 1;
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	$jobID = JobNum2ID ($job_cnt++);
	$mDiffFitJob{$fname1}{$fname2} = $jobID;
	print DAG "  <job id=\"$jobID\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
	print DAG "    <argument>";
	print DAG "-s ";
	print DAG "<filename file=\"$FitPlaneFname{$fname1}{$fname2}\"/>";
	print DAG "<filename file=\"$ProjFname{$fname1}\"/>";
	print DAG "<filename file=\"$ProjFname{$fname2}\"/>";
	print DAG "<filename file=\"$DiffFname{$fname1}{$fname2}\"/>";
	print DAG "<filename file=\"$template_fname\"/>";
	print DAG "</argument>\n";
	print DAG "    <uses file=\"$FitPlaneFname{$fname1}{$fname2}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$ProjFname{$fname1}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$AreaFname{$ProjFname{$fname1}}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$ProjFname{$fname2}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$AreaFname{$ProjFname{$fname2}}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$DiffFname{$fname1}{$fname2}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$AreaFname{$DiffFname{$fname1}{$fname2}}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$template_fname\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
	print DAG "  </job>\n";
	$count++;
    }
}
#
# Print XML "job" line for program to concatenate all of the fit plane 
# parameters output from the mDiffFit jobs into a single file.
#
$name = "mConcatFit";
$level--;
$count = 1;
$mConcatFitJob = JobNum2ID ($job_cnt++);
print DAG "  <job id=\"$mConcatFitJob\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
print DAG "    <argument>";
print DAG "<filename file=\"$fit_list\"/>";
print DAG "<filename file=\"$fit_planes\"/>";
print DAG " . ";
print DAG "</argument>\n";
print DAG "    <uses file=\"$fit_list\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
print DAG "    <uses file=\"$fit_planes\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	print DAG "    <uses file=\"fit.$FileNum{$fname1}.$FileNum{$fname2}.txt\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    }
}
print DAG "  </job>\n";
#
# Print XML "job" line for mBgModel job.
#
$name = "mBgModel";
$level--;
$mBgModelJob = JobNum2ID ($job_cnt++);
print DAG "  <job id=\"$mBgModelJob\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
print DAG "    <argument>";
print DAG "<filename file=\"$pimg_tbl\"/>";
print DAG "<filename file=\"$fit_planes\"/>";
print DAG "<filename file=\"$correction_planes\"/>";
print DAG "</argument>\n";
print DAG "    <uses file=\"$pimg_tbl\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
print DAG "    <uses file=\"$fit_planes\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
print DAG "    <uses file=\"$correction_planes\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
print DAG "  </job>\n";
#
# Print XML "job" lines for mBackground jobs.  Also, build a hash 
# linking filename to job ID for the mBackground job.
#
$name = "mBackground";
$level--;
$count = 1;
foreach $infile (@infiles) {
    $jobID = JobNum2ID ($job_cnt++);
    $mBgJob{$infile} = $jobID;
    print DAG "  <job id=\"$jobID\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
    print DAG "    <argument>";
    print DAG " -t ";
    print DAG "<filename file=\"$ProjFname{$infile}\"/>";
    print DAG "<filename file=\"$BgCorrFname{$infile}\"/>";
    print DAG "<filename file=\"$pimg_tbl\"/>";
    print DAG "<filename file=\"$correction_planes\"/>";
    print DAG "</argument>\n";
    print DAG "    <uses file=\"$ProjFname{$infile}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"$AreaFname{$ProjFname{$infile}}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"$pimg_tbl\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
    print DAG "    <uses file=\"$correction_planes\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"$BgCorrFname{$infile}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"$AreaFname{$BgCorrFname{$infile}}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "  </job>\n";
    $count++;
}
#
# Print XML "job" line for mImgtbl job(s) for background corrected images.  
# If the mosaic is large enough to be tiled, then this is done once for 
# each tile.
#
$name = "mImgtbl";
$level--;
$mImgtbl_count = 1;
for ($i = 0; $i < $NumTilesX; $i++) {
    for ($j = 0; $j < $NumTilesY; $j++) {
	$jobID = JobNum2ID ($job_cnt++);
	$mImgtblJob[$i][$j] = $jobID;
	print DAG "  <job id=\"$jobID\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name${mImgtbl_count}\" dv-version=\"$dv_version\">\n";
	print DAG "    <argument>";
	print DAG " . ";
	print DAG " -t ";
	print DAG "<filename file=\"$TiledCImgTbl[$i][$j]\"/>";
	print DAG "<filename file=\"$NewTiledCImgTbl[$i][$j]\"/>";
	print DAG "</argument>\n";
	print DAG "    <uses file=\"$TiledCImgTbl[$i][$j]\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
	print DAG "    <uses file=\"$NewTiledCImgTbl[$i][$j]\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	foreach $infile (@{$TiledFnames[$i][$j]}) {
	    print DAG "    <uses file=\"$infile\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	    print DAG "    <uses file=\"$AreaFname{$infile}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	}
	print DAG "  </job>\n";
	${mImgtbl_count}++;
    }
}
#
# Print XML "job" line for mAdd job(s) for background corrected images.  
# If the mosaic is large enough to be tiled, then this is done once for 
# each tile.
#
$name = "mAdd";
$level--;
$mAdd_count = 1;
for ($i = 0; $i < $NumTilesX; $i++) {
    for ($j = 0; $j < $NumTilesY; $j++) {
	$jobID = JobNum2ID ($job_cnt++);
	$mAddJob[$i][$j] = $jobID;
	print DAG "  <job id=\"$jobID\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name${mAdd_count}\" dv-version=\"$dv_version\">\n";
	print DAG "    <argument>";
	print DAG "-e ";
	print DAG "<filename file=\"$NewTiledCImgTbl[$i][$j]\"/>";
	print DAG "<filename file=\"$TiledTemplate[$i][$j]\"/>";
	print DAG "<filename file=\"$TiledOut[$i][$j]\"/>";
	print DAG "</argument>\n";
	print DAG "    <uses file=\"$NewTiledCImgTbl[$i][$j]\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	print DAG "    <uses file=\"$TiledTemplate[$i][$j]\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
	print DAG "    <uses file=\"$TiledOut[$i][$j]\" link=\"output\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
	print DAG "    <uses file=\"$AreaFname{$TiledOut[$i][$j]}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	foreach $infile (@{$TiledFnames[$i][$j]}) {
	    print DAG "    <uses file=\"$infile\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	    print DAG "    <uses file=\"$AreaFname{$infile}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	}
	print DAG "  </job>\n";
	${mAdd_count}++;
    }
}
#
# Print XML "job" line for mShrink job(s).  If the mosaic is large enough to be
# tiled, then this is done once for each tile.
#
$name = "mShrink";
$level--;
$count = 1;
for ($i = 0; $i < $NumTilesX; $i++) {
    for ($j = 0; $j < $NumTilesY; $j++) {
	$jobID = JobNum2ID ($job_cnt++);
	$mShrinkJob[$i][$j] = $jobID;
	print DAG "  <job id=\"$jobID\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
	print DAG "    <argument>";
	print DAG "<filename file=\"$TiledOut[$i][$j]\"/>";
	print DAG "<filename file=\"$ShrunkenTiledOut[$i][$j]\"/>";
	print DAG " $ShrinkFactor";
	print DAG "</argument>\n";
	print DAG "    <uses file=\"$TiledOut[$i][$j]\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
	print DAG "    <uses file=\"$ShrunkenTiledOut[$i][$j]\" link=\"output\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
	print DAG "  </job>\n";
	$count++;
    }
}
if ($NumTiles > 1) {
#
#   Print XML "job" line for mImgtbl job(s) for shrunken output images.  
#
    $name = "mImgtbl";
    $level--;
    $Shrunken_mImgtblJob = JobNum2ID ($job_cnt++);
    print DAG "  <job id=\"${Shrunken_mImgtblJob}\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name${mImgtbl_count}\" dv-version=\"$dv_version\">\n";
    print DAG "    <argument>";
    print DAG " . ";
    print DAG " -t ";
    print DAG "<filename file=\"${shrunken_tbl}\"/>";
    print DAG "<filename file=\"${new_shrunken_tbl}\"/>";
    print DAG "</argument>\n";
    print DAG "    <uses file=\"${shrunken_tbl}\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
    print DAG "    <uses file=\"${new_shrunken_tbl}\" link=\"output\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
     for ($i = 0; $i < $NumTilesX; $i++) {
        for ($j = 0; $j < $NumTilesY; $j++) {
            print DAG "    <uses file=\"$ShrunkenTiledOut[$i][$j]\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
            print DAG "    <uses file=\"$AreaFname{$ShrunkenTiledOut[$i][$j]}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	}
    }
    print DAG "  </job>\n";
#
#   Print XML "job" line for mAdd job for shrunken output images.  
#
    $name = "mAdd";
    $level--;
    $Shrunken_mAddJob = JobNum2ID ($job_cnt++);
    print DAG "  <job id=\"$Shrunken_mAddJob\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name${mAdd_count}\" dv-version=\"$dv_version\">\n";
    print DAG "    <argument>";
    print DAG "-n ";
    print DAG "-e ";
    print DAG "<filename file=\"${new_shrunken_tbl}\"/>";
    print DAG "<filename file=\"${shrunken_template_fname}\"/>";
    print DAG "<filename file=\"${shrunken_out_fname}\"/>";
    print DAG "</argument>\n";
    print DAG "    <uses file=\"${new_shrunken_tbl}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    print DAG "    <uses file=\"${shrunken_template_fname}\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
    print DAG "    <uses file=\"${shrunken_out_fname}\" link=\"output\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
    print DAG "    <uses file=\"$AreaFname{$shrunken_out_fname}\" link=\"output\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
    for ($i = 0; $i < $NumTilesX; $i++) {
        for ($j = 0; $j < $NumTilesY; $j++) {
            print DAG "    <uses file=\"$ShrunkenTiledOut[$i][$j]\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
            print DAG "    <uses file=\"$AreaFname{$ShrunkenTiledOut[$i][$j]}\" link=\"input\" dontRegister=\"true\" dontTransfer=\"true\" temporaryHint=\"tmp\"/>\n";
	}
    }
    print DAG "  </job>\n";
}
#
# Print XML "job" line for mJPEG job.
#
$name = "mJPEG";
$level--;
$count = 1;
$mJPEG_Job = JobNum2ID ($job_cnt++);
print DAG "  <job id=\"${mJPEG_Job}\" name=\"$name\" version=\"$MontageVersion\" level=\"$level\" dv-name=\"$name$count\" dv-version=\"$dv_version\">\n";
print DAG "    <argument>";
print DAG " -ct ";
print DAG "1 ";
print DAG "-gray ";
print DAG "<filename file=\"${shrunken_out_fname}\"/>";
print DAG " -1.5s ";
print DAG "60s ";
print DAG "2 ";
print DAG "-out ";
print DAG "<filename file=\"${shrunken_jpeg_fname}\"/>";
print DAG "</argument>\n";
print DAG "    <uses file=\"${shrunken_out_fname}\" link=\"input\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
print DAG "    <uses file=\"${shrunken_jpeg_fname}\" link=\"output\" dontRegister=\"false\" dontTransfer=\"false\"/>\n";
print DAG "  </job>\n";
#
# Print XML "child/parent" lines for mDiffFit jobs.
#
print DAG "<!-- part 3: list of control-flow dependencies (empty for single jobs) -->\n";
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	print DAG "  <child ref=\"$mDiffFitJob{$fname1}{$fname2}\">\n";
	print DAG "    <parent ref=\"$mProjJob{$fname1}\"/>\n";
	print DAG "    <parent ref=\"$mProjJob{$fname2}\"/>\n";
	print DAG "  </child>\n";
    }
}
#
# Print XML "child/parent" lines for job to concatenate all of the fit plane 
# paramter output from the mDiffFit jobs into a single file.
#
print DAG "  <child ref=\"$mConcatFitJob\">\n";
foreach $fname1 (keys %olaps) {
    foreach $fname2 (keys %{$olaps{$fname1}}) {
	print DAG "    <parent ref=\"$mDiffFitJob{$fname1}{$fname2}\"/>\n";
    }
}
print DAG "  </child>\n";
#
# Print XML "child/parent" lines for mBgModel job.
#
print DAG "  <child ref=\"$mBgModelJob\">\n";
print DAG "    <parent ref=\"$mConcatFitJob\"/>\n";
print DAG "  </child>\n";
#
# Print XML "child/parent" lines for mBackground jobs.
#
foreach $infile (@infiles) {
    print DAG "  <child ref=\"$mBgJob{$infile}\">\n";
    print DAG "    <parent ref=\"$mProjJob{$infile}\"/>\n";
    print DAG "    <parent ref=\"$mBgModelJob\"/>\n";
    print DAG "  </child>\n";
}
#
# Print XML "child/parent" lines for mImgtbl job(s) for background 
# corrected images.
#
for ($i = 0; $i < $NumTilesX; $i++) {
    for ($j = 0; $j < $NumTilesY; $j++) {
	print DAG "  <child ref=\"$mImgtblJob[$i][$j]\">\n";
	foreach $infile (@{$TiledFnames[$i][$j]}) {
	    print DAG "    <parent ref=\"$mBgJob{$InFname{$infile}}\"/>\n";
	}
	print DAG "  </child>\n";
    }
}
#
# Print XML "child/parent" lines for mAdd job(s) for background 
# corrected images.
#
for ($i = 0; $i < $NumTilesX; $i++) {
    for ($j = 0; $j < $NumTilesY; $j++) {
	print DAG "  <child ref=\"$mAddJob[$i][$j]\">\n";
	print DAG "    <parent ref=\"$mImgtblJob[$i][$j]\"/>\n";
	foreach $infile (@{$TiledFnames[$i][$j]}) {
	    print DAG "    <parent ref=\"$mBgJob{$InFname{$infile}}\"/>\n";
	}
	print DAG "  </child>\n";
    }
}
#
# Print XML "child/parent" lines for mShrink job(s).
#
for ($i = 0; $i < $NumTilesX; $i++) {
    for ($j = 0; $j < $NumTilesY; $j++) {
	print DAG "  <child ref=\"$mShrinkJob[$i][$j]\">\n";
	print DAG "    <parent ref=\"$mAddJob[$i][$j]\"/>\n";
	print DAG "  </child>\n";
    }
}
if ($NumTiles > 1) {
#
#   Print XML "child/parent" lines for mImgtbl job for shrunken output images.
#
    print DAG "  <child ref=\"${Shrunken_mImgtblJob}\">\n";
    for ($i = 0; $i < $NumTilesX; $i++) {
        for ($j = 0; $j < $NumTilesY; $j++) {
 	    print DAG "    <parent ref=\"$mShrinkJob[$i][$j]\"/>\n";
        }
    }
    print DAG "  </child>\n";
#
#   Print XML "child/parent" lines for mAdd job(s) for shrunken output images.
#
    print DAG "  <child ref=\"$Shrunken_mAddJob\">\n";
    print DAG "    <parent ref=\"$Shrunken_mImgtblJob\"/>\n";
    for ($i = 0; $i < $NumTilesX; $i++) {
        for ($j = 0; $j < $NumTilesY; $j++) {
	    print DAG "    <parent ref=\"$mShrinkJob[$i][$j]\"/>\n";
	}
    }
    print DAG "  </child>\n";
}
#
# Print XML "child/parent" lines for mJPEG job.
#
print DAG "  <child ref=\"${mJPEG_Job}\">\n";
if ($NumTiles > 1) {
    print DAG "    <parent ref=\"${Shrunken_mAddJob}\"/>\n";
}
else {
    print DAG "    <parent ref=\"$mShrinkJob[0][0]\"/>\n";
}
print DAG "  </child>\n";
#
# Print closing XML </adag> tag and terminate normally.
#
print DAG "</adag>\n";
close (DAG);
PrintLog ("Finished DAG processing for survey \"$survey\", location \"$MosaicCenter\", width \"$MosaicWidth\", height \"$MosaicHeight\", and band \"$band\"");
exit 0;
#
# Print a line of text with a time stamp.
#
sub PrintLog {
    my ($text) = @_;
    my $now = qx(date); chomp($now);
    print "$now : $text\n";
}
#
# Subroutine to show the usage syntax for this script.
#
sub ShowUsage {
    PrintLog ("Usage: $0 survey MosaicCenter MosaicWidth MosaicHeight band workdir email [suffix]");
}
#
# Subroutine to get location of Montage binaries used in this script and 
# ensure that they exist and are executable.  The MONTAGE environment variable
# must be set to specify the directory containing the Montage executables.
#
sub GetMontageBin {
    my ($MontageVersion, $mArchiveList, $mMakeHdr, $mOverlaps, $mDAGTbls,
	$mTileHdr, $mSubset) = @_;
    my $MontageEnv = "MONTAGE";
    my $bindir = $ENV{$MontageEnv}; # Montage bin directory.
    if (defined $bindir) {
	if ($bindir =~ /\/Montage_v([\d\.]+)\/bin$/) {
	    $$MontageVersion = $1;
	    my $err = 0;
	    $$mArchiveList = "$bindir/mArchiveList";
	    if (! -x $$mArchiveList) {
		PrintLog ("Couldn't execute $$mArchiveList!");
		$err = 1;
	    }
	    $$mMakeHdr = "$bindir/mMakeHdr";
	    if (! -x $$mMakeHdr) {
		PrintLog ("Couldn't execute $$mMakeHdr!");
		$err = 1;
	    }
	    $$mOverlaps = "$bindir/mOverlaps";
	    if (! -x $$mOverlaps) {
		PrintLog ("Couldn't execute $$mOverlaps!");
		$err = 1;
	    }
	    $$mDAGTbls = "$bindir/mDAGTbls";
	    if (! -x $$mDAGTbls) {
		PrintLog ("Couldn't execute $$mDAGTbls!");
		$err = 1;
	    }
	    $$mTileHdr = "$bindir/mTileHdr";
	    if (! -x $$mTileHdr) {
		PrintLog ("Couldn't execute $$mTileHdr!");
		$err = 1;
	    }
	    $$mSubset = "$bindir/mSubset";
	    if (! -x $$mSubset) {
		PrintLog ("Couldn't execute $$mSubset!");
		$err = 1;
	    }
	    if ($err) {
		PrintLog ("Please check that the Montage binaries exist in the directory specified by the $MontageEnv environment variable.");
		PrintLog ("$MontageEnv is currently set to $bindir.");
		exit 3;
	    }
	}
	else {
	    PrintLog ("The Montage binaries must exist in a subdirectory of the form Montage_v###/bin (examples: /usr/Montage_v2.2/bin or /usr/Montage_v2.2.1/bin)\n");
	    exit 4;
	}
    }
    else {
	PrintLog ("You must set the $MontageEnv environment variable to point to the directory containing the Montage executables before running this script.");
	exit 2;
    }
}
#
# Subroutine to build a list of input files, a hash linking filename 
# to a file number and another pair of hashes linking filename to projected 
# filename and background corrected filename, respectively.  Also, constructs
# a hash linking all filenames to corresponding area filenames.
#
sub ConstructFileInfo {
    my ($img_tbl_fullpath, $pimg_tbl_fullpath, $cimg_tbl_fullpath,
	$infiles, $ProjFname, $BgCorrFname, $InFname, $AreaFname, 
	$FileNum) = @_;
    my (@pnames, @cnames, $cnt, $fname, $pname, $cname);
    @$infiles = GetFilesFromTable ($img_tbl_fullpath);
    @pnames = GetFilesFromTable ($pimg_tbl_fullpath);
    if (@$infiles != @pnames) {
	PrintLog ("Cannot proceed because number of files in $pimg_tbl_fullpath does not match number of files in $img_tbl_fullpath!");
	exit 5;
    }
    @cnames = GetFilesFromTable ($cimg_tbl_fullpath);
    if (@$infiles != @cnames) {
	PrintLog ("Cannot proceed because number of files in $cimg_tbl_fullpath does not match number of files in $img_tbl_fullpath!");
	exit 5;
    }
    $cnt = 0;
    foreach $fname (@$infiles) {
	$pname = $pnames[$cnt];
	$cname = $cnames[$cnt];
	$$AreaFname{$fname} = GetAreaFname($fname);
	$$AreaFname{$pname} = GetAreaFname($pname);
	$$AreaFname{$cname} = GetAreaFname($cname);
	$$ProjFname{$fname} = $pname;
	$$BgCorrFname{$fname} = $cname;
	$$InFname{$cname} = $fname;
	$$FileNum{$fname} = $cnt + 1;
	$cnt++;
    }
}
#
# Subroutine to get input files.
#
sub GetFilesFromTable {
    my ($img_tbl_fullpath) = @_;
    my ($FieldCnt, $FnameCol, @FieldNames, @FileNames, $FileCnt, @fields);

    open (IN, "$img_tbl_fullpath") or die "Couldn't open $img_tbl_fullpath for reading: $!";
    $FnameCol = -1;
    $FileCnt = 0;
    while (<IN>) {
	chomp;
	if (/^\|/) {
#
#           ASSERT: Line is a header line.  Search for filename field column.
#
	    if ($FnameCol < 0) {
		@FieldNames = split /\s*\|\s*/;
		shift @FieldNames;
		$FieldCnt = 0;
		do {
		    if (($FieldNames[$FieldCnt] eq "file") ||
			($FieldNames[$FieldCnt] eq "fname")) {
			$FnameCol = $FieldCnt;
		    }
		    $FieldCnt++;
		} while (($FieldCnt < @FieldNames) && ($FnameCol < 0));
	    }
	}
	elsif (!/^\\/) { 
#
#       Line doesn't start with a backslash or vertical bar, so it must be a 
#       data line, not a header line.  Make sure we were able to detect which
#       column the filenames are in by now, and store the filename.
#
	    if ($FnameCol < 0) {
		PrintLog ("Could not extract filenames from $img_tbl_fullpath!");
		exit 4;
	    }
	    else {
		@fields = split;
		$FileNames[$FileCnt++] = $fields[$FnameCol];
	    }
	}
    }
    close (IN);
    return @FileNames;
}
#
# Subroutine to build a hash-of-hash flagging overlapping filenames.
#
sub GetOverlaps {
    my ($diffs_tbl_fullpath, $olaps) = @_;
    my ($fname1, $fname2);
    open (IN, "$diffs_tbl_fullpath") or die "Couldn't open $diffs_tbl_fullpath for reading: $!";
    while (<IN>) {
	chomp;
#	print "checking $_\n";
	if (!(/^\|/)) { #skip header lines
	    (undef, undef, $fname1, $fname2) = split;
	    $$olaps{$fname1}{$fname2} = 1;
#	    print "added $fname1,$fname2\n";
	}
    }
    close (IN);
}
#
# Subroutine to determine image size from a template header file.
#
sub GetImageSize {
    my ($template_fname_fullpath, $naxis1, $naxis2) = @_;
    my ($key, $val);
    open (TEMPLATE, "$template_fname_fullpath") or die "Couldn't open $template_fname_fullpath for reading $!";
    $$naxis1 = $$naxis2 = 0;
    while (<TEMPLATE>) {
	chomp;
	($key, $val) = split /\s*=\s*/;
	if ($key eq "NAXIS1") {
	    $$naxis1 = $val;
	    if ($$naxis2) {
		last;
	    }
	}
	elsif ($key eq "NAXIS2") {
	    $$naxis2 = $val;
	    if ($$naxis1) {
		last;
	    }
	}
    }
    close (TEMPLATE);
}
#
# Subroutine to print header lines for XML file.
#
sub XML_Header {
    print DAG <<EOF;
<?xml version="1.0" encoding="UTF-8"?>
EOF
    $now = qx(date); chomp($now);
    
    print DAG "<!-- generated: $now -->\n";
    print DAG "<!-- generated by: Montage DAG service -->\n";
    print DAG <<EOF;
<adag xmlns="http://www.griphyn.org/chimera/DAX" 
      xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" 
      xsi:schemaLocation="http://www.griphyn.org/chimera/DAX http://www.griphyn.org/chimera/dax-1.8.xsd" 
      count="1" index="0" name="test">
EOF
}
#
# Subroutine to convert a job number to a job ID string.
#
sub JobNum2ID {
    my ($jobnum) = @_;
    my $jobID = sprintf "ID%06d", $jobnum;
    return $jobID;
}
#
# Subroutine to convert from a file name to a corresponding area file name.
#
sub GetAreaFname {
    my ($fname) = @_;
    my (@parts, $AreaFname);
    @parts = split /\./, $fname;
    if (@parts == 1) { 
	$AreaFname = "${fname}_area";
    }
    else {
	$parts[-2] .= "_area";
	$AreaFname = join '.', @parts;
    }
    return $AreaFname;
}
#
# Subroutine to pad string $val up to length $width characters using $padchar
# as the padding character.
#
sub PadField {
    my ($val, $width, $padchar) = @_;
    my ($len, $pad);
    my $padval = $val;
    $len = length "$val";
    $pad = $width - $len;
    if ($pad > 0) {
	$HalfPad = sprintf ("%d", $pad / 2);
	$padval = $padchar x $HalfPad.$padval.$padchar x ($pad-$HalfPad);
    }
    return $padval;
}
#
# Subroutine to shrink (in a subsampling sense) a template header file.
#
sub ShrinkTemplate {
    my ($InTemplateFname, $ShrinkFactor, $ShrinkTemplateFname) = @_;
    open (IN, "$InTemplateFname") or die "Couldn't open $InTemplateFname for reading: $!";
    $count = 0;
    while (<IN>) {
	chomp;
	($key, $val) = split /\s*=\s*/;
	if ($key && $val) {
	    $TemplateVals{$key} = $val;
	    $keys[$count++] = $key;
	}
    }
    close (IN);
    $NormCRPIX1 = ($TemplateVals{"CRPIX1"} - 0.5) / $TemplateVals{"NAXIS1"};
    $NormCRPIX2 = ($TemplateVals{"CRPIX2"} - 0.5) / $TemplateVals{"NAXIS2"};
    $TemplateVals{"NAXIS1"} = sprintf ("%.0f", $TemplateVals{"NAXIS1"} / $ShrinkFactor);
    $TemplateVals{"NAXIS2"} = sprintf ("%.0f", $TemplateVals{"NAXIS2"} / $ShrinkFactor);
    $TemplateVals{"CDELT1"} *= $ShrinkFactor;
    $TemplateVals{"CDELT2"} *= $ShrinkFactor;
    $TemplateVals{"CRPIX1"} = $TemplateVals{"NAXIS1"} * $NormCRPIX1 + 0.5;
    $TemplateVals{"CRPIX2"} = $TemplateVals{"NAXIS2"} * $NormCRPIX2 + 0.5;
    open (OUT, ">$ShrinkTemplateFname") or die "Couldn't open $ShrinkTemplateFname for writing: $!";
    foreach $key (@keys) {
	print OUT "$key = $TemplateVals{$key}\n";
    }
    print OUT "END\n";
    close (OUT);
}

